/* Copyright (C) 1998, Jake Hamby.  All rights reserved.

  This file is part of Aladdin Ghostscript.

  Aladdin Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author
  or distributor accepts any responsibility for the consequences of using it,
  or for whether it serves any particular purpose or works at all, unless he
  or she says so in writing.  Refer to the Aladdin Ghostscript Free Public
  License (the "License") for full details.

  Every copy of Aladdin Ghostscript must include a copy of the License,
  normally in a plain ASCII text file named PUBLIC.  The License grants you
  the right to copy, modify and redistribute Aladdin Ghostscript, but only
  under certain conditions described in the License.  Among other things, the
  License requires that the copyright notice and this notice be preserved on
  all copies.
  

  gsfront.cpp

  Shared library front-end to Ghostscript.  Simulates traditional "gs"
  executable, but opens a window for the native BeOS GUI driver

  History
  1998 Initial port by Jake Hamby
  2000 Ported version 7 by Sander Stoks (see comment below)
  2004 Ported version 8 by Michael Pfeiffer (laplace@users.sourceforge.net)

  Note by Sander Stoks:
  This is mostly Jake Hamby's work.  All I did was modify it to use
  ArgvReceived() and RefsReceived() properly, so you can set this front-end
  to be the preferred app for application/postscript files and have it
  launch with the file opened when you double-click a PS file. 
  This is only of limited use since I left the "next page" stuff
  unimplemented.  A "proper" front end would open a small window with
  backward/forward buttons.  Maybe I'll write one on a rainy Sunday afternoon
  (and although I have plenty of those here in London, don't hold your breath :-) 
*/


#include <stdio.h>
#include <string.h>

#include <Alert.h>
#include <Application.h>
#include <Autolock.h>
#include <Beep.h>
#include <Entry.h>
#include <File.h>
#include <FindDirectory.h>
#include <MenuItem.h>
#include <Path.h>
#include <PopUpMenu.h>
#include <Screen.h>
#include <ScrollView.h>

typedef void *HWND;

extern "C" {
#include "gsdll.h"
void gsdll_draw(unsigned char *device, BView *view, BRect dest, BRect src);
}

// Prototypes
long gsloop(void *foo);

// Constants
/* Default dots-per-inch (be sure to choose an even number!) */
#define INITIAL_RESOLUTION ((int)96.0)
#define INITIAL_WIDTH ((int)(INITIAL_RESOLUTION * 85 / 10 + 1))
#define INITIAL_HEIGHT ((int)(INITIAL_RESOLUTION * 11 + 1))

static const char* kPreferencesFile = "ghostscript_settings";
static const char* kWindowLeft = "window_left";
static const char* kWindowTop = "window_top";
static const int kWindowTabHeight = 24;
static const int kWindowBorderSize = 4;

// Local class declarations

class Preferences {
public:
	Preferences();
	virtual ~Preferences();
	bool GetPath(BPath* path);

	BPoint GetWindowPosition()              { return fWindowPosition; }
	void SetWindowPosition(BPoint position) { fWindowPosition = position; }
	
private:
	BPoint fWindowPosition;
};


class PSView : public BView {
public:
	PSView(BRect area);
	void Draw(BRect area);
	void Pulse();
	void UpdateScrollbar(orientation direction, float viewSize, float pageSize);
	void UpdateScrollbars();
	void SetPageSize(int width, int height);
	
	void ScrollTo(BPoint point);

	BPoint GetMouseScreenPosition();

	void ShowPopUpMenu(BPoint point);
	void MouseDown(BPoint point);
	void MouseMoved(BPoint point, uint32 transit, const BMessage *msg);
	void MouseUp(BPoint point);
	
	void MouseWheelChanged(BMessage* msg);

	void MessageReceived(BMessage* msg);
	
	
private:
	int         fPageWidth;
	int         fPageHeight;
	// fields used for dragging support
	bool        fDragView;
	BPoint      fLastMousePosition;
};

class PSWindow : public BWindow {
public:
	PSWindow(BRect frame, const char *title, window_type type, ulong flags, Preferences* prefs);
	bool QuitRequested();
	void MessageReceived(BMessage* msg);
	
	void GsDllSync();
	void GsDllPage();
	void GsDllSetPageSize(int width, int height);
	void GsStopUpdate();
	
	void FrameMoved(BPoint position);
	void FrameResized(float width, float height);	

	PSView* GetView() { return fPSView; }
	
	
private:
	PSView      *fPSView;
	BScrollView *fScrollView;
	Preferences* fPreferences;
};

class PSApplication : public BApplication {
public:
    PSApplication(const char *sig);
	virtual ~PSApplication();
	
    void AboutRequested();
    void MessageReceived(BMessage *message);
    bool QuitRequested();
    void RefsReceived(BMessage *message);
    void ReadyToRun();


	PSWindow* GetWindow() { return fWindow; }
	Preferences* GetPreferences() { return fPreferences; }

private:
	Preferences *fPreferences;
    PSWindow    *fWindow;
	thread_id   fGhostscriptThread;
};

// Private message types
#define SHOW_FILE_PANEL	'SPNL'
#define NEXT_PAGE		'NXPG'

// The BeOS device to draw
unsigned char *gBeDevice = NULL;

// more global variables
PSApplication *gApplication;
int           gArgc;
char          **gArgv;
bool          gRunInTerminal;


// Implementation of Preferences
Preferences::Preferences()
	: fWindowPosition(15, kWindowTabHeight)
{
	BPath path;
	if (GetPath(&path)) {
		BFile file(path.Path(), B_READ_ONLY);
		BMessage data;
		if (file.InitCheck() == B_OK &&
			data.Unflatten(&file) == B_OK) {
			
			// read preferences
			data.FindFloat(kWindowLeft, &fWindowPosition.x);
			data.FindFloat(kWindowTop, &fWindowPosition.y);
		}
	}
}

Preferences::~Preferences()
{
	BPath path;
	if (GetPath(&path)) {
		// write preferences
		BMessage data;
		data.AddFloat(kWindowLeft, fWindowPosition.x);
		data.AddFloat(kWindowTop, fWindowPosition.y);
		
		BFile file(path.Path(), B_WRITE_ONLY | B_CREATE_FILE | B_ERASE_FILE);
		if (file.InitCheck() == B_OK) {
			data.Flatten(&file);
		}
	}
}

bool Preferences::GetPath(BPath* path)
{
	return find_directory(B_COMMON_SETTINGS_DIRECTORY, path) == B_OK &&
		path->Append(kPreferencesFile) == B_OK;
}


// Implementation of PSWindow

PSWindow::PSWindow(BRect frame, const char *title, window_type type, ulong flags, Preferences* prefs) 
	: BWindow(frame, title, type, flags)
	, fPreferences(prefs)
{
	AddShortcut('a', B_COMMAND_KEY, new BMessage(B_ABOUT_REQUESTED));
	
	frame.OffsetTo(B_ORIGIN);
	frame.right -= B_V_SCROLL_BAR_WIDTH;
	frame.bottom -= B_H_SCROLL_BAR_HEIGHT;
	fPSView = new PSView(frame);
	
	fScrollView = new BScrollView("Scroll View", fPSView, B_FOLLOW_ALL,
			0, true, true, B_FANCY_BORDER);
	fScrollView->ScrollBar(B_HORIZONTAL)->SetRange(0, 611);
	fScrollView->ScrollBar(B_VERTICAL)->SetRange(0, 791);

	AddChild(fScrollView);
	fPSView->MakeFocus();
}

bool PSWindow::QuitRequested() 
{
	be_app->PostMessage(B_QUIT_REQUESTED);
	return true;
}

void PSWindow::MessageReceived(BMessage* msg)
{
	if (msg->what == B_ABOUT_REQUESTED) {
		be_app->PostMessage(B_ABOUT_REQUESTED);
	}
}

void PSWindow::GsDllSync()
{
	BAutolock autoLock(this);
	if (!autoLock.IsLocked()) return;
	// Use pulse to update the view while the page is being rendered
	fPSView->Invalidate();
	SetPulseRate(1000000/10);
}

void PSWindow::GsDllPage()
{
	BAutolock autoLock(this);
	if (!autoLock.IsLocked()) return;
	// Stop pulse
	SetPulseRate(0);
	fPSView->Invalidate();
}

void PSWindow::GsStopUpdate()
{
	BAutolock autoLock(this);
	if (autoLock.IsLocked()) {
		SetPulseRate(0);
	}
}

// Resize window but stay inside screen bounds
void PSWindow::GsDllSetPageSize(int width, int height)
{
	BAutolock autoLock(this);
	
	fPSView->SetPageSize(width, height);
	
	BRect screensize;
	BScreen screen;
	if (screen.IsValid()) {
		screensize = screen.Frame();
	} else {
		screensize = BRect(0, 0, 640, 480);
	}
	
	// scroll bars should not overlap page
	width += (int)B_V_SCROLL_BAR_WIDTH;
	height += (int)B_H_SCROLL_BAR_HEIGHT;
	
	BRect frame(Frame());
	// stay inside screen
	if (frame.left + width - kWindowBorderSize > screensize.right) {
		width = (int)(screensize.right - frame.left - kWindowBorderSize);
	}
	if (frame.top + height - kWindowBorderSize > screensize.bottom) {
		height = (int)(screensize.bottom - frame.top - kWindowBorderSize);
	}
	ResizeTo(width, height);
}

void PSWindow::FrameMoved(BPoint position)
{
	if (fPreferences != NULL) {
		fPreferences->SetWindowPosition(position);
	}
}

void PSWindow::FrameResized(float width, float height)
{
	fPSView->UpdateScrollbars();
}

// Implementation of PSView

PSView::PSView(BRect frame) 
	: BView(frame, "PSView", B_FOLLOW_ALL_SIDES, B_WILL_DRAW | B_PULSE_NEEDED) 
	, fPageWidth(INITIAL_WIDTH)
	, fPageHeight(INITIAL_HEIGHT)
{
	SetViewColor(B_TRANSPARENT_COLOR);
}

void PSView::Draw(BRect area) 
{
	const rgb_color kWhite = {255, 255, 255, 0};
	SetDrawingMode(B_OP_COPY);
	SetHighColor(kWhite);

	if(gBeDevice) {
		// fill space around page with white color
		if (area.left < 0) {
			FillRect(BRect(area.left, area.top, -1, area.bottom));
		}
		if (fPageWidth <= area.right) {
			FillRect(BRect(fPageWidth, area.top, area.right, area.bottom));
		}
		if (area.top < 0) {
			FillRect(BRect(area.left, area.top, area.right, -1));
		}
		if (fPageHeight <= area.bottom) {
			FillRect(BRect(area.left, fPageHeight, area.right, area.bottom));			
		}
		// paint page
		gsdll_lock_device(gBeDevice, 1);
		gsdll_draw(gBeDevice, this, area, area);
		gsdll_lock_device(gBeDevice, 0);
	} else {
		FillRect(area);
	}
}

void PSView::Pulse()
{
	Invalidate();
}

void PSView::SetPageSize(int width, int height)
{
	fPageWidth = width;
	fPageHeight = height;
}

void PSView::UpdateScrollbar(orientation direction, float viewSize, float pageSize) 
{
	BScrollBar* scrollBar = ScrollBar(direction);
	if (scrollBar != NULL) {
		float outsideSize = pageSize - viewSize;
		if (outsideSize < 0.0) {
			outsideSize = 0.0;
		}
		
		scrollBar->SetRange(0, outsideSize);
		scrollBar->SetProportion(viewSize / pageSize);
		scrollBar->SetSteps(viewSize / 10.0, viewSize);
	}
}

void PSView::UpdateScrollbars() 
{
	BRect bounds(Bounds());
	UpdateScrollbar(B_HORIZONTAL, bounds.Width(), fPageWidth);
	UpdateScrollbar(B_VERTICAL, bounds.Height(), fPageHeight);
}

// scroll restricted within page size
void PSView::ScrollTo(BPoint point) 
{
	BRect bounds(Bounds());
	float xMax = fPageWidth - bounds.Width();
	float yMax = fPageHeight - bounds.Height();
	
	if (xMax < 0) xMax = 0;	
	if (yMax < 0) yMax = 0;
	
	if (point.x < 0) {
		point.x = 0;
	} else if (point.x > xMax) {
		point.x = xMax;
	}
	
	if (point.y < 0) {
		point.y = 0;
	} else if (point.y > yMax)  {
		point.y = yMax;
	}
		
	BView::ScrollTo(point); 
}

BPoint PSView::GetMouseScreenPosition()
{
	uint32 buttons;
	BPoint position;
	GetMouse(&position, &buttons, false);
	ConvertToScreen(&position);
	return position;
}

void PSView::ShowPopUpMenu(BPoint point)
{
	BPopUpMenu* menu = new BPopUpMenu("PopUpMenu");
	menu->SetAsyncAutoDestruct(true);

	BMessage* msg;
	BMenuItem* i;

	// About menu item
	msg = new BMessage(B_ABOUT_REQUESTED);
	i = new BMenuItem("About", msg);
	i->SetTarget(this);
	menu->AddItem(i);

	point -= BPoint(10, 10);
	menu->Go(point, true, false, false);	
}

void PSView::MouseDown(BPoint point)
{
	uint32 buttons;
	GetMouse(&point, &buttons);
	if (buttons & B_PRIMARY_MOUSE_BUTTON) {
		MakeFocus();
		SetMouseEventMask(B_POINTER_EVENTS);
		fLastMousePosition = GetMouseScreenPosition();
		fDragView = true;	
	} else if (buttons & B_SECONDARY_MOUSE_BUTTON) {
		ShowPopUpMenu(GetMouseScreenPosition());
	}
}

void PSView::MouseMoved(BPoint point, uint32 transit, const BMessage *msg)
{
	if (fDragView) {
		BPoint position(GetMouseScreenPosition());
		BPoint delta(fLastMousePosition - position);
		ScrollBy(delta.x, delta.y);
		fLastMousePosition = position;
	}
}

void PSView::MouseUp(BPoint point)
{
	MouseMoved(point, 0, NULL);
	fDragView = false;	
}

void PSView::MouseWheelChanged(BMessage* msg)
{
	float scrollDistance = 25;
	
	// set scroll distance to 20% of the minimum of screen width and height
	BScreen screen;
	if (screen.IsValid()) {
		BRect frame(screen.Frame());
		if (frame.right < frame.bottom) {
			scrollDistance = frame.right;
		} else {
			scrollDistance = frame.bottom;
		}
		scrollDistance *= 0.20;
	}
	
	float dy, dx;
	if (msg->FindFloat("be:wheel_delta_y", &dy) == B_OK && dy != 0.0) {
		bool down = dy > 0;
		// intelliMouse driver uses command key to simulate wheel_detla_x!
		if ((modifiers() & (B_COMMAND_KEY | B_OPTION_KEY))) {
			// Zoom(!down); // zoom in / out	
		} else if ((modifiers() & (B_SHIFT_KEY | B_CONTROL_KEY))) {
			// next/previous page
			// MoveToPage(mCurrentPage + (down ? 1 : -1));
		} else {
			ScrollBy(0, scrollDistance * (down ? 1 : -1));
		}
	}
	if (msg->FindFloat("be:wheel_delta_x", &dx) == B_OK && dx != 0.0) {
		bool right = dx > 0;
		ScrollBy(scrollDistance * (right ? 1 : -1), 0);
	}
}

void PSView::MessageReceived(BMessage* msg)
{
//	if (msg->what == B_MOUSE_WHEEL_CHANGED) {
		// MouseWheelChanged(msg);
//	} else {
		BView::MessageReceived(msg);
//	}
}

// Implementation of PSApplication

PSApplication::PSApplication(const char *sig) : BApplication(sig) 
{
	fPreferences = new Preferences();
	
	BScreen myscreen;
	BRect screensize;
	if (myscreen.IsValid()) {
		screensize = myscreen.Frame();
	} else {
		screensize = BRect(0,0,640,480);
	}
	
	BPoint position(15, kWindowTabHeight);
	if (fPreferences != NULL) {
		position = fPreferences->GetWindowPosition();
	}
	BRect frame(position.x, position.y, 
		screensize.right - kWindowBorderSize, screensize.bottom - kWindowBorderSize);
		
    fWindow = new PSWindow(frame, "Ghostscript", B_DOCUMENT_WINDOW, 0, fPreferences);
        		
	fWindow->Run();
}

PSApplication::~PSApplication() {
	delete fPreferences;
	fPreferences = NULL;
}

void PSApplication::ReadyToRun ()
{
	fGhostscriptThread = spawn_thread(gsloop, "GS Loop", B_NORMAL_PRIORITY, (void *)GetWindow()->GetView());
	resume_thread(fGhostscriptThread);
}

void PSApplication::AboutRequested() {
   BAlert *alert;

   alert = new BAlert("", "AFPL Ghostscript 8.14 (2004-06-14)\n\n"
	"Copyright (C) 2004 artofcode LLC, Benicia, CA.  All rights reserved.\n"
	"This software comes with NO WARRANTY: see the file PUBLIC for details.\n\n"
	"BeOS port by Jake Hamby, Sander Stoks and Michael Pfeiffer.\n\n"
	"Please send all comments to:\n"
	"    laplace@users.sourceforge.net",
   	"OK", NULL, NULL, B_WIDTH_AS_USUAL, B_INFO_ALERT);
   alert->Go();
}

void PSApplication::MessageReceived(BMessage *message)
{
	switch (message->what) {
		case SHOW_FILE_PANEL:
//			RunFilePanel();         // Use default values
			fprintf(stderr, "Open file...\n");
			break;
        case NEXT_PAGE:
			fprintf(stderr, "Next page...\n");
			break;
		default:
			BApplication::MessageReceived(message);
            break;
	}
}

bool PSApplication::QuitRequested() {
	kill_thread(fGhostscriptThread);  // Kill Ghostscript thread if still running
	return true;                      // and exit app
}

void PSApplication::RefsReceived(BMessage *message)
{
	gRunInTerminal = false;
	BPath path;
	entry_ref ref;
	if (message->FindRef ("refs", &ref) == B_OK) {
		gArgc = 2;
		gArgv[0] = new char[16];
		strcpy (gArgv[0], "gs");
		path.SetTo (&ref);
		gArgv[1] = new char[strlen(path.Path()) + 1];
		strcpy (gArgv[1], path.Path());
	}
}

// Only used to open and close BWindow
int
gsdll_callback(int message, char *str, unsigned long data)
{
	static bool child_added = 0;
    char *p;
    switch (message) {
	    case GSDLL_STDIN:
	        p = fgets(str, data, stdin);
	        if (p)
	           return strlen(str);
	        else
	           return 0; 
	    case GSDLL_STDOUT:
	        if (str != (char *)NULL)
				fwrite(str, 1, data, stderr);
	        return data;
	    case GSDLL_DEVICE:
			//fprintf(stderr, "GSDLL_DEVICE called\n");
			gBeDevice = (unsigned char *)str;
			if (data && !child_added) {	// open window
				child_added = 1;
				gApplication->GetWindow()->Lock();
				gApplication->GetWindow()->Show();
				gApplication->GetWindow()->Unlock();
			} else {	// close window
				// do nothing right now
			}
			break;
	    case GSDLL_SYNC:
	        //fprintf(stderr,"Callback: SYNC %p %ld\n", str, data);
	        gApplication->GetWindow()->GsDllSync();
	        break;
	    case GSDLL_PAGE:
	        //fprintf(stderr,"Callback: PAGE %p %ld\n", str, data);
	        gApplication->GetWindow()->GsDllPage();
	        break;
	    case GSDLL_SIZE:
	        //fprintf(stderr,"Callback: SIZE %p width=%d height=%d\n", str,
	        //    (int)(data & 0xffff), (int)((data>>16) & 0xffff) );
			{
			int width = (int)(data & 0xffff);
			int height = (int)((data>>16) & 0xffff);
			gApplication->GetWindow()->GsDllSetPageSize(width, height);
			}
	        break;
	    case GSDLL_POLL:
	        //fprintf(stderr,"Callback: POLL %p\n", str);
	        return 0; /* no error */
	    default:
	        //fprintf(stderr,"Callback: Unknown message=%d\n",message);
	        break;
    }
    return 0;
}

void stop_automatic_invalidation()
{
	gApplication->GetWindow()->GsStopUpdate();
}

long gsloop(void *psview)
{
	char buf[256];
    
    int code;
    code = gsdll_init(gsdll_callback, NULL, gArgc, gArgv);
	// fprintf(stderr, "gsdll_init returns %d\n", code);

	if (code == 0) {
    	code = gsdll_execute_begin();
    }

	// fprintf(stderr, "gsdll_execute_begin returns %d\n", code);
    if (code==0) {
		stop_automatic_invalidation();
		fprintf(stderr, "GS>");
		fflush(stderr);
		while(1)
		{
		    if(!fgets(buf, sizeof(buf)-1, stdin))
				break;

			code = gsdll_execute_cont(buf, strlen(buf));
//			fprintf(stderr, "gsdll_execute returns %d\n", code);

			if (code < 0)
		    	break;
			
			stop_automatic_invalidation();
			fprintf(stderr, "GS>");
			fflush(stderr);
		}

		if (!code) {
			code = gsdll_execute_end();
		}
		
		code = gsdll_exit();
		// fprintf(stderr, "gsdll_exit returns %d\n", code);
	}

	be_app->PostMessage(B_QUIT_REQUESTED);
    return 0;
}

int main (int32 argc, char *argv[])
{
	gArgc = argc;
	gArgv = argv;
    PSApplication app("application/x-vnd.jehamby.gs");
    gApplication = &app;	// so global functions can find the app
    
    gRunInTerminal = true;
    
    app.Run();
    return 0;
}

