/* BeOS driver for Ghostscript using a BBitmap for buffering.
 * This driver has been heavily rewritten from the previous version.
 *
 * Copyright 1998, Jake Hamby
 * Send all comments, bug fixes, etc. to:  jehamby@lightside.com
 *
 * Anti-Aliasing drivers: bealpha2, bealpht2, bealpha4 and bealpht4 
 * Copyright 2000, Michael Pfeiffer
 * Email: laplace@users.sourceforge.net
 *
 * Description: bealpha2 and bealpha4 anti-alias text and graphic. 
 * bealpht2 and bealpht4 anti-alias text only.
 * 
 * History
 * 1998 initial port by Jake Hamby
 * 2000 ported version 7 by Sander Stoks 
 *      anti-aliasing driver by Michael Pfeiffer
 * 2004 ported version 8, by Michael Pfeiffer
 *		documented, refactored and improved AA device
 *      (about 50% faster for certain kind of PS files e.g. tiger.eps) 
 *      
 *
 * To do (eventually): 
 * -  Creating and deleting BBitmap object is an inefficient operation!
 *    If BE_ALPHA_BITMAP_CACHE then
 *    - check if temporary list is faster on SMP then _cached_bitmap
 *    - on cache overflow try to reuse a large enough bitmap instead
 *      of deleting bitmaps from cache and creating a new one
 *    - on cache overflow delete large bitmaps first
 * -  Find optimal values for (see gdevbe.h):
 *    - BE_ALPHA_HASHTABLE_SIZE
 *    - BE_ALPHA_MAX_BITMAPS
 *    - BE_ALPHA_REMOVE_BITMAPS
 *    - BE_ALPHA_MAX_SIZE
 *    and an optimal hash function HASH_FUNC in this file.
 * -  Set the bits per pixel for the internal bitmap with a 
 *    parameter (e.g. gs -dBitsPerPixel={8,16,32}).
 * -  Limit the number of bitmaps in the mono bitmap cache.
 * -  Let the GS interpreter be interruptible (return a negative number
 *    in the device-callback functions).
 */

// set PROFILE to 1 to enable profiling
#define PROFILE 0
// set TRACE_COPY_CALLS to 1 to enable tracing of copy mono and alpha function calls
#define TRACE_COPY_CALLS 0

// define macro if certain functions should be inlined
//#define INLINE inline
// define macro if inlining should be disabled
#define INLINE 

#include <View.h>
#include <Window.h>
#include <Bitmap.h>

typedef void* HWND;       // For old DLL interface

extern "C" {
	void gsdll_draw(unsigned char *device, BView *view, BRect dest, BRect src);

	#define __PROTOTYPES__    // C++ needs prototypes, of course
	#include "gx.h"			  // for gx_bitmap; includes std.h 
	#include "math_.h"
	#include "memory_.h"
	#include "gserrors.h"
	#include "gsparam.h"
	#include "gsstruct.h"
	#include "gxdevice.h"
	#include "gsdevice.h"
	#include "gdevbe.h"
	#include "gsdll.h"

	// Intel does name mangling on variables, so define these here.
	extern gx_device_be gs_bealpha2_device;	
	extern gx_device_be gs_bealpha4_device;	
	extern gx_device_be gs_bealpht2_device;	
	extern gx_device_be gs_bealpht4_device;	
	extern gx_device_be gs_beos_device;	
	extern HWND hwndtext;		        
} // extern "C"

// Macros
#if 0
#pragma mark *********** Macros ***************
#endif
// The hash function
#define BE_ALPHA_HASH(key) ((key) % BE_ALPHA_HASHTABLE_SIZE)
// Color compare function
#define SAME_COLOR(c1, c2) (*((uint32*)&(c1)) == *((uint32*)&(c2)))


// Types
#if 0
#pragma mark *********** Types ***************
#endif
// type to convert values between rgb_color and gx_color_index
typedef union {
	rgb_color rgb;
	gx_color_index index;
} colorcast_t;

// The BBitmap cache entry
// The bitmaps are stored in a linked list
typedef struct bmapcache_s {
	gx_bitmap_id id;          // the id of the bitmap
	struct bmapcache_s *next; // the next entry in the linked list
	BBitmap bitmap;           // the BBitmap
	
	// constructor
	bmapcache_s(
		gx_bitmap_id id, 
		BRect bounds, 
		color_space mode) 
	: 
		id(id), 
		next(NULL), 
		bitmap(bounds, mode) // construct the BBitmap
	{ 
		// nothing to do
	}
} bmapcache;


// The BBitmap alpha cache struct 
typedef struct bmap_alpha_cache_s {
	gx_bitmap_id id; // the id of the bitmap stored in this entry
	rgb_color color; // the color of this bitmap
	long size;       // the size of this bitmap in 4 bytes
	struct bmap_alpha_cache_s *of_next;   // the next entry in the overflow list
	struct bmap_alpha_cache_s *of_prev;   // the previous entry in the overflow list
	struct bmap_alpha_cache_s *mru_next;  // the next entry in the MRU list (most recently used)
	struct bmap_alpha_cache_s *mru_prev;  // the previous entry in the MRU list
	struct bmap_alpha_cache_s **ht_entry; // the hashtable entry with this bitmap 	
	BBitmap bitmap;

	// constructor	
	bmap_alpha_cache_s(
		struct bmap_alpha_cache_s **ht_entry, 
		gx_bitmap_id id, 
		rgb_color color, 
		int w, int h, 
		color_space mode) 
	:
		id(id), 
		color(color), 
		size(w*h), 
		of_next(NULL), 
		of_prev(NULL), 
		mru_next(NULL), 
		mru_prev(NULL), 
		ht_entry(ht_entry),
		bitmap(BRect(0, 0, w-1, h-1), mode) // construct BBitmap
	{ 
		// nothing to do
	}
} bmap_alpha_cache;

// Global variables
HWND hwndtext;               /* currently unused */

// alpha conversion table
// 2 bit alpha -> 8 bit alpha (= 255/((1 << 2) - 1) * alpha)
static unsigned char g_alpha2[4] = {0, 85, 170, 255};
// 4 bit alpha -> 8 bit alpha (= 255/((1 << 4) - 1) * alpha)
static unsigned char g_alpha4[16] = {0, 17, 34, 51, 68, 85, 102, 119, 136, 153, 170, 187, 204, 221, 238, 255};



// Procedures for non anti-aliased device (named beos)
#if 0
#pragma mark *********** GS Procedures ***************
#endif

private dev_proc_open_device(be_open);
private dev_proc_sync_output(be_sync);
private dev_proc_output_page(be_output_page);
private dev_proc_close_device(be_close);
private dev_proc_map_rgb_color(be_map_rgb_color);
private dev_proc_map_color_rgb(be_map_color_rgb);
private dev_proc_fill_rectangle(be_fill_rectangle);
private dev_proc_copy_mono(be_copy_mono);
#if 0
private dev_proc_copy_color(be_copy_color);
#endif
private dev_proc_draw_line(be_draw_line);

#if 0
// Not Implemented
private dev_proc_strip_tile_rectangle(be_strip_tile_rectangle);

private dev_proc_get_params(be_get_params);
private dev_proc_put_params(be_put_params);
#endif

// Procedures for anti-aliased devices (bealpha2, bealpha4, bealpht2, bealpht4)
private dev_proc_map_color_rgb(be_alpha_map_color_rgb);
private dev_proc_map_rgb_alpha_color(be_alpha_map_rgb_alpha_color);
private dev_proc_copy_alpha(be_alpha_copy_alpha);


#if 0
#pragma mark *********** Device Descriptors & Devices ***************
#endif
#if 0
#pragma mark ** beos **
#endif
// The device descriptor used by gs_beos_device
private gx_device_procs be_procs = {
	be_open,
	gx_default_get_initial_matrix,
	be_sync,
	be_output_page,
	be_close,
	be_map_rgb_color,
	be_map_color_rgb,
	be_fill_rectangle,
	NULL,			/* tile_rectangle */
	be_copy_mono,
	NULL,			/* copy_color */
	be_draw_line,
	NULL,			/* get_bits */
	NULL, // be_get_params,			/* get_params */
	NULL, // be_put_params,			/* put_params */
	NULL,			/* map_cmyk_color */
	NULL,			/* get_xfont_procs */
	NULL,			/* get_xfont_device */
	NULL,			/* map_rgb_alpha_color */
	gx_page_device_get_page_device,
	NULL,			/* get_alpha_bits */
	NULL,			/* copy_alpha */
	NULL,			/* get_band */
	NULL,			/* copy_rop */
	NULL,			/* fill_path */
	NULL,			/* stroke_path */
	NULL,			/* fill_mask */
	NULL,			/* fill_trapezoid */
	NULL,			/* fill_parallelogram */
	NULL,			/* fill_triangle */
	NULL,			/* draw_thin_line */
	NULL,			/* begin_image */
	NULL,			/* image_data */
	NULL,			/* end_image */
	NULL			/* be_strip_tile_rectangle */
};

/* The instance is public. */
gx_device_be gs_beos_device = {
	std_device_color_body(gx_device_be, &be_procs, "beos",
	  INITIAL_WIDTH, INITIAL_HEIGHT,
	  INITIAL_RESOLUTION, INITIAL_RESOLUTION,
	  24, 255, 256),
	{ 0 },			/* std_procs */
	-1,				/* mutex */
	NULL,			/* main bitmap */
	NULL,			/* offscreen view */
	NULL,			/* bmapcache ptr */
	0,				/* next bitmap to use */
};

/* ---------------------------------- */
/* ------------ bealpha2 ------------ */
/* ---------------------------------- */
#if 0
#pragma mark ** bealpha2 **
#endif

/* The device descriptor */
private gx_device_procs be_alpha2_procs = {
	be_open,
	gx_default_get_initial_matrix,
	be_sync,
	be_output_page,
	be_close,
	be_map_rgb_color,
	be_alpha_map_color_rgb, // be_map_color_rgb,
	be_fill_rectangle,
	NULL,			/* tile_rectangle */
	be_copy_mono,
	NULL,			/* copy_color */
	be_draw_line,
	NULL,			/* get_bits */
	NULL,			/* get_params */
	NULL,			/* put_params */
	NULL,			/* map_cmyk_color */
	NULL,			/* get_xfont_procs */
	NULL,			/* get_xfont_device */
	be_alpha_map_rgb_alpha_color,			/* map_rgb_alpha_color */
	gx_page_device_get_page_device,
	NULL,           /* get_alpha_bits */
	be_alpha_copy_alpha,			/* copy_alpha */
	NULL,			/* get_band */
	NULL,			/* copy_rop */
	NULL,			/* fill_path */
	NULL,			/* stroke_path */
	NULL,			/* fill_mask */
	NULL,			/* fill_trapezoid */
	NULL,			/* fill_parallelogram */
	NULL,			/* fill_triangle */
	NULL,			/* draw_thin_line */
	NULL,			/* begin_image */
	NULL,			/* image_data */
	NULL,			/* end_image */
	NULL			/* be_strip_tile_rectangle */
};

/* The instance is public. */
gx_device_be gs_bealpha2_device = {
    std_device_dci_alpha_type_body(gx_device_be, &be_alpha2_procs,
	"bealpha2", NULL,
	  INITIAL_WIDTH, INITIAL_HEIGHT,
	  INITIAL_RESOLUTION, INITIAL_RESOLUTION,
	3, 32, 255, 255, 256, 256, 2, 2),
	{ 0 },			/* std_procs */
	-1,				/* mutex */
	NULL,			/* main bitmap */
	NULL,			/* offscreen view */
	NULL,			/* bmapcache ptr */
	0,				/* next bitmap to use */
};


/* ---------------------------------- */
/* ------------ bealpha4 ------------ */
/* ---------------------------------- */
#if 0
#pragma mark ** bealpha4 **
#endif

/* The device descriptor */
private gx_device_procs be_alpha4_procs = {
	be_open,
	gx_default_get_initial_matrix,
	be_sync,
	be_output_page,
	be_close,
	be_map_rgb_color,
	be_alpha_map_color_rgb, // be_map_color_rgb,
	be_fill_rectangle,
	NULL,			/* tile_rectangle */
	be_copy_mono,
	NULL,			/* copy_color */
	be_draw_line,
	NULL,			/* get_bits */
	NULL,			/* get_params */
	NULL,			/* put_params */
	NULL,			/* map_cmyk_color */
	NULL,			/* get_xfont_procs */
	NULL,			/* get_xfont_device */
	be_alpha_map_rgb_alpha_color,			/* map_rgb_alpha_color */
	gx_page_device_get_page_device,
	NULL, // be_alpha4_get_alpha_bits,			/* get_alpha_bits */
	be_alpha_copy_alpha,			/* copy_alpha */
	NULL,			/* get_band */
	NULL,			/* copy_rop */
	NULL,			/* fill_path */
	NULL,			/* stroke_path */
	NULL,			/* fill_mask */
	NULL,			/* fill_trapezoid */
	NULL,			/* fill_parallelogram */
	NULL,			/* fill_triangle */
	NULL,			/* draw_thin_line */
	NULL,			/* begin_image */
	NULL,			/* image_data */
	NULL,			/* end_image */
	NULL			/* be_strip_tile_rectangle */
};

/* The instance is public. */
gx_device_be gs_bealpha4_device = {
    std_device_dci_alpha_type_body(gx_device_be, &be_alpha4_procs,
	"bealpha4", NULL,
	  INITIAL_WIDTH, INITIAL_HEIGHT,
	  INITIAL_RESOLUTION, INITIAL_RESOLUTION,
	3, 32, 255, 255, 256, 256, 4, 4),
	{ 0 },			/* std_procs */
	-1,				/* mutex */
	NULL,			/* main bitmap */
	NULL,			/* offscreen view */
	NULL,			/* bmapcache ptr */
	0,				/* next bitmap to use */
};

/* ---------------------------------- */
/* ------------ bealpht2 ------------ */
/* ---------------------------------- */
#if 0
#pragma mark ** bealpht2 **
#endif

/* The device descriptor */
private gx_device_procs be_alpht2_procs = {
	be_open,
	gx_default_get_initial_matrix,
	be_sync,
	be_output_page,
	be_close,
	be_map_rgb_color,
	be_alpha_map_color_rgb, // be_map_color_rgb,
	be_fill_rectangle,
	NULL,			/* tile_rectangle */
	be_copy_mono,
	NULL,			/* copy_color */
	be_draw_line,
	NULL,			/* get_bits */
	NULL,			/* get_params */
	NULL,			/* put_params */
	NULL,			/* map_cmyk_color */
	NULL,			/* get_xfont_procs */
	NULL,			/* get_xfont_device */
	be_alpha_map_rgb_alpha_color,			/* map_rgb_alpha_color */
	gx_page_device_get_page_device,
	NULL,			/* get_alpha_bits */
	be_alpha_copy_alpha,			/* copy_alpha */
	NULL,			/* get_band */
	NULL,			/* copy_rop */
	NULL,			/* fill_path */
	NULL,			/* stroke_path */
	NULL,			/* fill_mask */
	NULL,			/* fill_trapezoid */
	NULL,			/* fill_parallelogram */
	NULL,			/* fill_triangle */
	NULL,			/* draw_thin_line */
	NULL,			/* begin_image */
	NULL,			/* image_data */
	NULL,			/* end_image */
	NULL			/* be_strip_tile_rectangle */
};

/* The instance is public. */
gx_device_be gs_bealpht2_device = {
    std_device_dci_alpha_type_body(gx_device_be, &be_alpht2_procs,
	"bealpht2", NULL,
	  INITIAL_WIDTH, INITIAL_HEIGHT,
	  INITIAL_RESOLUTION, INITIAL_RESOLUTION,
	3, 32, 255, 255, 256, 256, 2, 1),
	{ 0 },			/* std_procs */
	-1,				/* mutex */
	NULL,			/* main bitmap */
	NULL,			/* offscreen view */
	NULL,			/* bmapcache ptr */
	0,				/* next bitmap to use */
};

/* ---------------------------------- */
/* ------------ bealpht4 ------------ */
/* ---------------------------------- */
#if 0
#pragma mark ** bealphat4 **
#endif

/* The device descriptor */
private gx_device_procs be_alpht4_procs = {
	be_open,
	gx_default_get_initial_matrix,
	be_sync,
	be_output_page,
	be_close,
	be_map_rgb_color,
	be_alpha_map_color_rgb, // be_map_color_rgb,
	be_fill_rectangle,
	NULL,			/* tile_rectangle */
	be_copy_mono,
	NULL,			/* copy_color */
	be_draw_line,
	NULL,			/* get_bits */
	NULL,			/* get_params */
	NULL,			/* put_params */
	NULL,			/* map_cmyk_color */
	NULL,			/* get_xfont_procs */
	NULL,			/* get_xfont_device */
	be_alpha_map_rgb_alpha_color,			/* map_rgb_alpha_color */
	gx_page_device_get_page_device,
	NULL, 			/* get_alpha_bits */
	be_alpha_copy_alpha,			/* copy_alpha */
	NULL,			/* get_band */
	NULL,			/* copy_rop */
	NULL,			/* fill_path */
	NULL,			/* stroke_path */
	NULL,			/* fill_mask */
	NULL,			/* fill_trapezoid */
	NULL,			/* fill_parallelogram */
	NULL,			/* fill_triangle */
	NULL,			/* draw_thin_line */
	NULL,			/* begin_image */
	NULL,			/* image_data */
	NULL,			/* end_image */
	NULL			/* be_strip_tile_rectangle */
};

/* The instance is public. */
gx_device_be gs_bealpht4_device = {
    std_device_dci_alpha_type_body(gx_device_be, &be_alpht4_procs,
	"bealpht4", NULL,
	  INITIAL_WIDTH, INITIAL_HEIGHT,
	  INITIAL_RESOLUTION, INITIAL_RESOLUTION,
	3, 32, 255, 255, 256, 256, 4, 1),
	{ 0 },			/* std_procs */
	-1,				/* mutex */
	NULL,			/* main bitmap */
	NULL,			/* offscreen view */
	NULL,			/* bmapcache ptr */
	0,				/* next bitmap to use */
};


// -----------------------------------------------------------------------------
// Bitmap Cache Functions
// -----------------------------------------------------------------------------

// Initialize the bitmap cache
private void
initialize_bitmap_cache(gx_device_be* bdev)
{
	bdev->_bmapcache_current = &(bdev->_bmapcache);
}

// flush
private void
flush_bitmap_cache(gx_device_be* bdev) 
{
	bmapcache_s *cur = bdev->_bmapcache;
	while(cur) {
		bmapcache_s *next = cur->next;
		delete cur;
		cur = next;
	}
	bdev->_bmapcache = NULL;
	bdev->_bmapcache_current = &(bdev->_bmapcache);
}

// store bitmap in bitmap cache
private void
add_to_bitmap_cache(gx_device_be* bdev, bmapcache_s* bmapc)
{
	*(bdev->_bmapcache_current) = bmapc;
	bdev->_bmapcache_current = &(bmapc->next);
}

// find bitmap in bitmap cache
private BBitmap*
find_bitmap(gx_device_be* bdev, gx_bitmap_id id)
{
	if(id != gx_no_bitmap_id) {
		bmapcache_s *current = bdev->_bmapcache;
		while(current) {
		    if(current->id == id) {
		    	return &(current->bitmap);
		    }
			current = current->next;
		}
	}
	return NULL;
}


// -----------------------------------------------------------------------------
// Alpha Bitmap Cache Functions
// -----------------------------------------------------------------------------

// Initialize cache
private void
be_alpha_initialize(gx_device_be* bdev)
{
	// initialize hash table
	for (int i = 0; i < BE_ALPHA_HASHTABLE_SIZE; i++) {
		bdev->_alpha_hashtable[i] = NULL;
	}
	// initialize MRU list
	bdev->_alpha_mru_first = NULL;
	bdev->_alpha_mru_last = NULL;
	// initialize run time statistics
	bdev->_alpha_num_bmaps = 0;
	bdev->_alpha_size = 0;
}

// Uninitialize cache
private void
be_alpha_uninitialize(gx_device_be* bdev)
{
	// remove bitmaps from hash table
	bmap_alpha_cache *current, *old;
	for (int i = 0; i < BE_ALPHA_HASHTABLE_SIZE; i++) {
		current = bdev->_alpha_hashtable[i];
		while (current) {
			old = current;
			current = current->of_next;
			delete old;
		}
		bdev->_alpha_hashtable[i] = NULL;
	}
		
	bdev->_alpha_mru_first = NULL;
	bdev->_alpha_mru_last  = NULL;
	bdev->_alpha_num_bmaps = 0;
	bdev->_alpha_size      = 0;
}

// Find bmapc in hash table
// Returns entry in hash table and bmapc if it is in overflow list
private INLINE bmap_alpha_cache **
be_alpha_find_bmap(
	gx_device_be *bdev, 
	gx_bitmap_id id, 
	rgb_color color, 
	bmap_alpha_cache **bmapc) 
{
	bmap_alpha_cache **ht_entry = &bdev->_alpha_hashtable[BE_ALPHA_HASH(id)];
	bmap_alpha_cache *current = *ht_entry;
	// find bitmap in overflow list
	while (current) {
	 	if ((current->id == id) && SAME_COLOR(current->color, color)) {
	 		break; // found
	 	}
		current = current->of_next;
	}
	*bmapc = current;
	return ht_entry;
}

// Insert bmapc at top of overflow list of hashtable entry ht_entry
private INLINE void 
be_alpha_ht_insert(
	gx_device_be *bdev, 
	bmap_alpha_cache **ht_entry, 
	bmap_alpha_cache *bmapc) 
{
	bmapc->of_next = *ht_entry;
	bmapc->of_prev = NULL;
	if (*ht_entry) {
		(*ht_entry)->of_prev = bmapc;
	}
	*ht_entry = bmapc;
}

// Remove bmapc from overflow list of hasttable entry ht_entry
private INLINE void 
be_alpha_ht_remove(
	gx_device_be *bdev, 
	bmap_alpha_cache **ht_entry, 
	bmap_alpha_cache *bmapc) 
{
	// is bmapc on top of overflow list?
	if (*ht_entry == bmapc) {
		*ht_entry = bmapc->of_next;
	}
	// update doubly linked overflow list
	if (bmapc->of_next) {
		bmapc->of_next->of_prev = bmapc->of_prev;
	}
	if (bmapc->of_prev) {
		bmapc->of_prev->of_next = bmapc->of_next;
	}
}

// Insert bmapc at top of MRU list. bmpac must not be in list already!
private INLINE void 
be_alpha_mru_insert(
	gx_device_be *bdev, 
	bmap_alpha_cache *bmapc) 
{
	// update doubly linked list
	bmapc->mru_next = bdev->_alpha_mru_first;
	bmapc->mru_prev = NULL;
	if (bdev->_alpha_mru_first) {
		// MRU list was not empty
		bdev->_alpha_mru_first->mru_prev = bmapc;
	} else { 
		// bmapc is first element in MRU list
		bdev->_alpha_mru_last = bmapc; 
	}
	// bmapc becomes top of MRU list	
	bdev->_alpha_mru_first = bmapc;
}

// Remove bmapc from MRU list
private INLINE void 
be_alpha_mru_remove(
	gx_device_be *bdev, 
	bmap_alpha_cache *bmapc) 
{
	if (bdev->_alpha_mru_first == bmapc) { 
		bdev->_alpha_mru_first = bmapc->mru_next;
	}
	if (bdev->_alpha_mru_last == bmapc) {
		bdev->_alpha_mru_last = bmapc->mru_prev;
	}
	if (bmapc->mru_next) {
		bmapc->mru_next->mru_prev = bmapc->mru_prev;
	}
	if (bmapc->mru_prev) {
		bmapc->mru_prev->mru_next = bmapc->mru_next;
	}
}

// Move bmapc on top of overflow list of hashtable entry ht_entry
private INLINE void 
be_alpha_update_of_list(
	gx_device_be *bdev, 
	bmap_alpha_cache **ht_entry, 
	bmap_alpha_cache *bmapc) 
{
	if (*ht_entry != bmapc) {
		be_alpha_ht_remove(bdev, ht_entry, bmapc);
		be_alpha_ht_insert(bdev, ht_entry, bmapc);
	}
}

// Move bmapc in front of mru list
private INLINE void 
be_alpha_update_mru_list(
	gx_device_be *bdev, 
	bmap_alpha_cache *bmapc) 
{
	if (bdev->_alpha_mru_first != bmapc) {
		be_alpha_mru_remove(bdev, bmapc);
		be_alpha_mru_insert(bdev, bmapc);
	}
}

// Move bmapc on top of overflow list and MRU list
private INLINE void 
be_alpha_move_to_top(
	gx_device_be *bdev, 
	bmap_alpha_cache **ht_entry, 
	bmap_alpha_cache *bmapc) 
{
	be_alpha_update_of_list(bdev, ht_entry, bmapc);
	be_alpha_update_mru_list(bdev, bmapc);
}

// Add bmapc to overflow list and MRU list
private INLINE void 
be_alpha_insert_bmap(
	gx_device_be *bdev, 
	bmap_alpha_cache **ht_entry, 
	bmap_alpha_cache *bmapc) 
{
	be_alpha_ht_insert(bdev, ht_entry, bmapc);
	be_alpha_mru_insert(bdev, bmapc);
	bdev->_alpha_num_bmaps ++;
	bdev->_alpha_size += bmapc->size;
}

// Remove bmapc from overflow list and MRU list
private INLINE void 
be_alpha_remove_bmap(
	gx_device_be *bdev, 
	bmap_alpha_cache **ht_entry, 
	bmap_alpha_cache *bmapc) 
{
	be_alpha_ht_remove(bdev, ht_entry, bmapc);
	be_alpha_mru_remove(bdev, bmapc);
	bdev->_alpha_num_bmaps --;
	bdev->_alpha_size -= bmapc->size;
}


// -----------------------------------------------------------------------------
// Device functions
// -----------------------------------------------------------------------------

/* Open the device. */
private int
be_open(gx_device *dev)
{
	// gx_device_be has been created for us
    gx_device_be *bdev = (gx_device_be *)dev;
    initialize_bitmap_cache(bdev);
	// create semaphore for locking
	bdev->_mutex = create_sem(1, "gsdll sem");
	
	if (dev->width == INITIAL_WIDTH)
		dev->width = (int)(8.5 * dev->x_pixels_per_inch);
	if (dev->height == INITIAL_HEIGHT)
		dev->height = (int)(11.0 * dev->y_pixels_per_inch);

	// initialize cached bitmap
	bdev->_cached_bitmap = NULL;
	bdev->_cached_width  = 0;
	bdev->_cached_height = 0;

	// setup BBitmap and BView
	BRect frame(0.0, 0.0, dev->width-1, dev->height-1);
	bdev->_bitmap = new BBitmap(frame, B_RGB_32_BIT, true);
	bdev->_view = new BView(frame, "bitmap view", B_FOLLOW_NONE, B_WILL_DRAW);
	bdev->_bitmap->AddChild(bdev->_view);
	bdev->_view->Window()->Lock();
	bdev->_view->SetBlendingMode(B_PIXEL_ALPHA, B_ALPHA_OVERLAY);

	be_alpha_initialize(bdev);

#if PROFILE
	// initialize profiler statistics
	bdev->_fillrect  = 0;
	bdev->_copymono  = 0;
	bdev->_drawline  = 0;
	bdev->_copyalpha = 0;
	bdev->_size      = 0; 
	bdev->_bitmaps   = 0;
#endif

	// Notify caller about new device
	(*pgsdll_callback)(GSDLL_DEVICE, (char *)dev, 1);
	// Notify caller about the page size
	(*pgsdll_callback)(GSDLL_SIZE, (char *)dev,
			(dev->width & 0xffff) +
			((ulong)(dev->height & 0xffff) << 16));
    return 0;
}

#if PROFILE
private void
print_profile_statistics(gx_device_be* bdev)
{
    fprintf(stderr, "fillrect=%d copymono=%d drawline=%d copyalpha=%d "
    	"bitmaps=%ld size=%ld\n",
    	bdev->_fillrect, 
    	bdev->_copymono, 
    	bdev->_drawline, 
    	bdev->_copyalpha,
   		bdev->_bitmaps, 
   		bdev->_size);

    fprintf(stderr, "alpha_num_bmaps=%d alpha_size=%ld\n",
   		bdev->_alpha_num_bmaps, 
   		bdev->_alpha_size);

}
#endif

/* Close the device. */
private int
be_close(gx_device *dev)
{
    gx_device_be *bdev = (gx_device_be *)dev;

    // wait until bitmap is not being used by caller
    gsdll_lock_device((unsigned char*)dev, 1);

    // notify caller that device has been closed
	(*pgsdll_callback)(GSDLL_DEVICE, (char *)dev, 0);
	gsdll_lock_device((unsigned char*)dev, 0);


	// complete all drawing operations
	bdev->_view->Sync();
	
	// delete bitmap; deletes view too
	delete bdev->_bitmap;
	
	// delete cached bitmap
	delete bdev->_cached_bitmap;
	
	// delete monochrome bitmap cache
	flush_bitmap_cache(bdev);

	be_alpha_uninitialize(bdev);

	delete_sem(bdev->_mutex);

#if PROFILE
	print_profile_statistics(bdev);
#endif

    return 0;
}

/* Map a color to index.  */
private gx_color_index
be_map_rgb_color(
	register gx_device *dev,
	const gx_color_value rgb[])
{
    colorcast_t colorcast;
	colorcast.rgb.red   = rgb[0] >> 8;
	colorcast.rgb.green = rgb[1] >> 8;
	colorcast.rgb.blue  = rgb[2] >> 8;
	colorcast.rgb.alpha = 0;
	return colorcast.index;
}


/* Map a "device color" back to r-g-b. */
private int
be_map_color_rgb(
	register gx_device *dev, 
	gx_color_index color,
	gx_color_value prgb[3])
{
	colorcast_t colorcast;
	
	colorcast.index = color;
	prgb[0] = colorcast.rgb.red << 8;
	prgb[1] = colorcast.rgb.green << 8;
	prgb[2] = colorcast.rgb.blue << 8;
	return 0;
}

// Synchronize the display with the commands already given 
private int
be_sync(register gx_device *dev)
{
    gx_device_be *bdev = (gx_device_be *)dev;
    
    bdev->_view->Sync();
    
    return (*pgsdll_callback)(GSDLL_SYNC, (char *)dev, 0);
}

// Display the output.
int
be_output_page(gx_device *dev, int copies, int flush)
{
    gx_device_be *bdev = (gx_device_be *)dev;
    
    bdev->_view->Sync();
    
	int ret = (*pgsdll_callback)(GSDLL_PAGE, (char *)dev, 0);

#if PROFILE
	print_profile_statistics(bdev);
#endif

    return ret;
}

// Fill a rectangle with a color.
private int
be_fill_rectangle(register gx_device *dev,
		 int x, int y, int w, int h, gx_color_index color)
{
#if PROFILE
    bigtime_t start = system_time();
#endif

    gx_device_be *bdev = (gx_device_be *)dev;
    
	colorcast_t colorcast;
	colorcast.index = color;
	bdev->_view->SetDrawingMode(B_OP_COPY);
	bdev->_view->SetHighColor(colorcast.rgb);
	bdev->_view->FillRect(BRect(x, y, w+x-1, y+h-1));

	if(x <= 0 && y <= 0 && w >= bdev->width && h >= bdev->height) {
		bdev->_view->Sync();
		// Just cleared the screen so clear the bitmap cache 
		flush_bitmap_cache(bdev);
	}

#if PROFILE
	bdev->_fillrect += (int)(system_time() - start);
#endif

	return 0;
}

/* Copy a monochrome bitmap. */
private int
be_copy_mono(register gx_device *dev,
	    const byte *base, int sourcex, int raster, gx_bitmap_id id,
	    int x, int y, int w, int h,
	    gx_color_index zero, gx_color_index one)
{
#if TRACE_COPY_CALLS
	fprintf(stderr, "*+-m\t%d\t%d\t%d\t%ld\t%ld\n",
		(int)id,
		w, h,
		(unsigned long)zero,
		(unsigned long)one);
#endif

#if PROFILE
    bigtime_t start = system_time();
#endif

    gx_device_be *bdev = (gx_device_be *)dev;

    if(raster > 0) {
		colorcast_t colorcast;
		
		// set high color
		colorcast.index = one;
		bdev->_view->SetHighColor(colorcast.rgb);
		
		// set low color
		if(zero == gx_no_color_index) {
			// transparent pixels
			bdev->_view->SetDrawingMode(B_OP_OVER);
		} else {
			colorcast.index = zero;
			bdev->_view->SetLowColor(colorcast.rgb);
			bdev->_view->SetDrawingMode(B_OP_COPY);
		}
	
		BBitmap *bitmap = NULL;
		bitmap = find_bitmap(bdev, id);
		bool bitmapInCache = bitmap != NULL;
		
		BRect from(sourcex, 0, sourcex + w - 1, h - 1);
		BRect to(x, y, x + w - 1, y + h - 1);
		
		bmapcache_s *bmapc = NULL;
		if(!bitmapInCache) {
			// create new bitmap
			bmapc = new bmapcache_s(id, BRect(0.0, 0.0, (raster << 3) - 1, h - 1), B_MONOCHROME_1_BIT);
		    bmapc->bitmap.SetBits(base, raster * h, 0, B_MONOCHROME_1_BIT);
			bitmap = &(bmapc->bitmap);
			
		    if (id != gx_no_bitmap_id) {
		    	// store bitmap in cache
		    	add_to_bitmap_cache(bdev, bmapc);
				bdev->_view->DrawBitmapAsync(bitmap, from, to);
				bitmapInCache = true;
			}
		}
		
		if (bitmapInCache) {
			bdev->_view->DrawBitmapAsync(bitmap, from, to);
		} else {
			bdev->_view->DrawBitmap(bitmap, from, to);
			delete bmapc;
		}
    }
    
#if PROFILE
	bdev->_copymono += (int)(system_time() - start);
#endif

    return 0;
}

/* Copy a color bitmap. */
/* Note:  I couldn't find any PS file which uses this function, so I left
 * it unimplemented for now.  -jeh
 */
#if 0
private int
be_copy_color(register gx_device *dev,
	     const byte *base, int sourcex, int raster, gx_bitmap_id id,
	     int x, int y, int w, int h)
{
    fprintf(stderr, "copy color!\n");
    return 0;
}
#endif

/* Draw a line */
private int
be_draw_line(register gx_device *dev,
  	     int x0, int y0, int x1, int y1, 
  	     gx_color_index color)
{
#if PROFILE
    bigtime_t start = system_time();
#endif

    gx_device_be *bdev = (gx_device_be *)dev;
	colorcast_t colorcast;
	colorcast.index = color;
	bdev->_view->SetDrawingMode(B_OP_COPY);
	bdev->_view->SetHighColor(colorcast.rgb);
	bdev->_view->StrokeLine(BPoint(x0, y0), BPoint(x1, y1));

#if PROFILE
	bdev->_drawline += (int)(system_time() - start);
#endif

    return 0;
}

/* Tile a rectangle. */
/* Note:  Because the Be GUI always has at least 256 colors, and because
 *   this function is only used for halftoning, I left it unimplemented for
 *   now.  -jeh
 */
#if 0
private int
be_strip_tile_rectangle(register gx_device *dev, const gx_strip_bitmap *tiles,
		       int x, int y, int w, int h,
		       gx_color_index zero, gx_color_index one,
		       int px, int py)
{
    fprintf(stderr, "Striptile rectangle\n");
    return 0;
}
#endif

// Lock to make DLL threadsafe
#if defined(__POWERPC__)
__declspec(export)
#endif
int gsdll_lock_device(unsigned char *dev, int flag) {
	gx_device_be *bdev = (gx_device_be *)dev;
	
    if (flag) {
    	acquire_sem(bdev->_mutex);
    	return 1;
    }
    
    release_sem(bdev->_mutex);
    return 0;
}

// Draw to on-screen BView 
#if defined(__POWERPC__)
__declspec(export)
#endif
void gsdll_draw(unsigned char *dev, BView *view, BRect dest, BRect src) {
	gx_device_be *bdev = (gx_device_be *)dev;
	view->SetDrawingMode(B_OP_COPY);
	view->DrawBitmapAsync(bdev->_bitmap, src, dest);
}

#if 0
// TODO Use an argument to specify whether to use anti-aliasing or not
// and the the bit depth of the alpha channel.
int be_get_params(gx_device *dev, gs_param_list *plist) {
	if (plist) {
		gs_param_enumerator_t iterator;
		gs_param_key_t key;
		param_init_enumerator(&iterator);
		while (0 == param_get_next_key(plist, &iterator, &key)) {
			if (key.data) {
				fprintf(stderr, "%*.*s %d\n", (int)key.size, (int)key.size, key.data);
			}
		}
	}
	return 0;
}

int be_put_params(gx_device *dev, gs_param_list *plist) {	
	if (plist) {
		gs_param_enumerator_t iterator;
		gs_param_key_t key;
		param_init_enumerator(&iterator);
		while (0 == param_get_next_key(plist, &iterator, &key)) {
			if (key.data) {
				fprintf(stderr, "%*.*s %d\n", (int)key.size, (int)key.size, key.data);
			}
		}
	
		int bpp;
		param_read_int(plist, "BitsPerPixel", &bpp);
		long l;
		param_read_long(plist, "BitsPerPixel", &l);
	}
	return 0;
}
#endif


/* Anti-aliasing Devices: bealpha2, bealpha4, bealpht2, bealpht4 */

// Color conversion from gx_color_index to gx_color_value
private int
be_alpha_map_color_rgb(
	gx_device * dev, 
	gx_color_index color,
	gx_color_value prgb[3])
{
    colorcast_t colorcast;
	colorcast.index = color;
	prgb[0] = colorcast.rgb.red   << 8;
	prgb[1] = colorcast.rgb.green << 8;
	prgb[2] = colorcast.rgb.blue  << 8;
	return 0;	
}

// Color conversion from gx_color_values to index
private gx_color_index
be_alpha_map_rgb_alpha_color(
	gx_device * dev,
	gx_color_value r, 
	gx_color_value g, 
	gx_color_value b, 
	gx_color_value alpha)
{
    colorcast_t colorcast;
	colorcast.rgb.red   = r >> 8;
	colorcast.rgb.green = g >> 8;
	colorcast.rgb.blue  = b >> 8;
	colorcast.rgb.alpha = alpha >> 8;
	return colorcast.index;
}


private void
make_room_in_alpha_cache(gx_device_be* bdev)
{
	bmap_alpha_cache *bmapc;
	// Remove bitmaps from cache if required
	if ((bdev->_alpha_num_bmaps >= BE_ALPHA_MAX_BITMAPS) || (bdev->_alpha_size >= BE_ALPHA_MAX_SIZE)) {
		bdev->_view->Sync();
		// Do we need to remove more bitmaps?
		bmapc = bdev->_alpha_mru_last;
		if (bmapc && 
			((bdev->_alpha_num_bmaps >= BE_ALPHA_MAX_BITMAPS) || (bdev->_alpha_size >= BE_ALPHA_MAX_SIZE))) {
			// Remove the least recently used bitmaps
			do {
				be_alpha_remove_bmap(bdev, bmapc->ht_entry, bmapc);
				delete bmapc;
				bmapc = bdev->_alpha_mru_last;
			} while (bmapc && 
				((bdev->_alpha_num_bmaps >= (BE_ALPHA_MAX_BITMAPS - BE_ALPHA_REMOVE_BITMAPS)) || 
				(bdev->_alpha_size >= BE_ALPHA_MAX_SIZE)));
		}
	}
} 

// 2 alpha bits per pixel  (4 alpha values per byte in base)
private void
alpha2_to_bitmap(const unsigned char *base, int raster, int w, int h, BBitmap* bitmap, colorcast_t colorcast)
{
	unsigned char *bmp = (unsigned char*)bitmap->Bits();
	int32 bpr = bitmap->BytesPerRow();
	const unsigned char *data = base;
	int32 w4 = w/4;
	int32 rest = w & 3;
	
	for (int32 line = h; line; line --) {
		// start of line
		const unsigned char *src = data; 
		rgb_color *dst = (rgb_color*)bmp; 
		
		// next line
		data += raster;
		bmp += bpr;
		
		// fill bitmap
		for (int32 column = w4; column; column --) {
			colorcast.rgb.alpha = g_alpha2[(*src) >> 6];
			*dst = colorcast.rgb;
			dst++;
			
			colorcast.rgb.alpha = g_alpha2[((*src) >> 4) & 0x03];
			*dst = colorcast.rgb;
			dst++;
			
			colorcast.rgb.alpha = g_alpha2[((*src) >> 2) & 0x03];
			*dst = colorcast.rgb;
			dst++;
			
			colorcast.rgb.alpha = g_alpha2[(*src) & 0x03];
			*dst = colorcast.rgb;
			dst++; src++;
		}
		
		switch (rest) {
			case 1:
				colorcast.rgb.alpha = g_alpha2[(*src) >> 6];
				*dst = colorcast.rgb;
				break;

			case 2:
				colorcast.rgb.alpha = g_alpha2[(*src) >> 6];
				*dst = colorcast.rgb;
				dst++;
				
				colorcast.rgb.alpha = g_alpha2[((*src) >> 4) & 0x03];
				*dst = colorcast.rgb;
				break;
				
			case 3:
				colorcast.rgb.alpha = g_alpha2[(*src) >> 6];
				*dst = colorcast.rgb;
				dst++;
				
				colorcast.rgb.alpha = g_alpha2[((*src) >> 4) & 0x03];
				*dst = colorcast.rgb;
				dst++;
				
				colorcast.rgb.alpha = g_alpha2[((*src) >> 2) & 0x03];
				*dst = colorcast.rgb;
				break;
		}
	}
}


// 4 alpha bits per pixel (2 alpha values per byte in base)
private void
alpha4_to_bitmap(const unsigned char *base, int raster, int w, int h, BBitmap* bitmap, colorcast_t colorcast)
{
	unsigned char *bmp = (unsigned char*)bitmap->Bits();
	int32 bpr = bitmap->BytesPerRow();
	const unsigned char *data = base;
	int32 w2 = w/2;
	bool rest = (w & 1) > 0;
	
	for (int32 line = h; line; line --) {
		// start of line
		const unsigned char *src = data; 
		rgb_color *dst = (rgb_color*)bmp;
		
		// next line
		data += raster;
		bmp += bpr;
		
		// fill bitmap
		for (int32 column = w2; column; column --) {
			colorcast.rgb.alpha = g_alpha4[(*src) >> 4];
			*dst = colorcast.rgb;
			dst++;
			
			colorcast.rgb.alpha = g_alpha4[(*src) & 0x0F];
			*dst = colorcast.rgb;
			src++; dst++;
		}
	
		if (rest) {
			colorcast.rgb.alpha = g_alpha4[(*src) >> 4];
			*dst = colorcast.rgb;
		}
	}
}

// return a bitmap that is large enough (if necessary delete the cached bitmap
// and create a new one)
private BBitmap*
get_cached_bitmap(gx_device_be* bdev, int w, int h)
{
	BBitmap* bitmap = bdev->_cached_bitmap;
	if (bitmap == NULL || (bdev->_cached_width < w) || (bdev->_cached_height < h)) {
		delete bitmap;
		bitmap = new BBitmap(BRect(0.0, 0.0, w-1, h-1), B_RGBA32);
		bdev->_cached_bitmap = bitmap;
		bdev->_cached_width = w;
		bdev->_cached_height = h;
	}
	return bitmap;
}

// Copy alpha mask
private int
be_alpha_copy_alpha(
	gx_device * dev, 
	const unsigned char *base, 
	int sourcex,
	int raster, 
	gx_bitmap_id id, 
	int x, int y, int w, int h,
	gx_color_index color, 
	int depth)
{
#if TRACE_COPY_CALLS
	fprintf(stderr, "*+-a\t%d\t%d\t%d\t%ld\t%d\n",
		(int)id,
		w, h,
		(unsigned long)color,
		depth);
#endif

#if PROFILE
    bigtime_t start = system_time();
#endif

    gx_device_be *bdev = (gx_device_be *)dev;

    if(raster > 0) {
		colorcast_t colorcast;
		colorcast.index = color;
	
		// We have to swap red and blue values, so we can use colorcast.rgb to write directly to memory
		unsigned char swap = colorcast.rgb.red; 
		colorcast.rgb.red = colorcast.rgb.blue; 
		colorcast.rgb.blue = swap;
	
		BBitmap *bitmap = NULL;
	
		bmap_alpha_cache **ht_entry = NULL;
		bmap_alpha_cache *bmapc;
		
		
		// look in alpha cache
		if (id != gx_no_bitmap_id) {
			ht_entry = be_alpha_find_bmap(bdev, id, colorcast.rgb, &bmapc);
			if (bmapc != NULL) { 
				// cache hit 
				bitmap = &(bmapc->bitmap);
				be_alpha_move_to_top(bdev, ht_entry, bmapc);
			}
		}
	
		if(bitmap == NULL) {
			// Create a bitmap in the color 'color'; take alpha channel from 'base'
			if (ht_entry) {
			#if 0
				// if cache is full
				//   if cache has reuseable bitmap
				//      take bitmap, move to top, sync
				//   else
				//      delete some bitmaps from cache
				// else
				//   create a new bitmap
			#else
				make_room_in_alpha_cache(bdev);
				bmapc = new bmap_alpha_cache_s(ht_entry, id, colorcast.rgb, w, h, B_RGBA32);
				bitmap = &(bmapc->bitmap); 
				be_alpha_insert_bmap(bdev, ht_entry, bmapc);
			#endif
			} else {
				// Because bitmaps are drawn asynchronously wait for completion
				// otherwise we could delete or overwrite a currently being drawn bitmap!
				bdev->_view->Sync();
				bitmap = get_cached_bitmap(bdev, w, h);
			}

			// convert the alpha mask to BBitmap
			if (depth == 2) { 
				alpha2_to_bitmap(base, raster, w, h, bitmap, colorcast);
			} else if (depth == 4) { 
				alpha4_to_bitmap(base, raster, w, h, bitmap, colorcast);
			} else { 
				// should not reach here!
			}
		} // bitmap == NULL

		bdev->_view->SetDrawingMode(B_OP_ALPHA);

		if (bitmap != NULL) {		
			BRect from(sourcex, 0, sourcex + w - 1, h - 1);
			BRect to(x, y, x + w - 1, y + h - 1);
			bdev->_view->DrawBitmapAsync(bitmap, from, to);
		} // bitmap != NULL
    } // raster > 0
    
#if PROFILE
	bdev->_copyalpha += (int)(system_time() - start);
	bdev->_bitmaps++; bdev->_size += w * h;
#endif
    
    return 0;
}

